import multiprocessing as mp
import pickle

from reho.model.master_problem import *
from reho.model.postprocessing.KPIs import *
from reho.model.postprocessing.building_scale_network_builder import *
from reho.paths import *

__doc__ = """
File for constructing and solving the optimization problem.
"""


class REHO(MasterProblem):
    """
    Performs the single or multi-objective optimization.

    Parameters are inherited from ``MasterProblem``.

    See also
    --------
    reho.model.master_problem.MasterProblem
    """

    def __init__(self, qbuildings_data, units, grids, parameters=None, set_indexed=None, cluster=None, method=None, scenario=None, solver="highs", DW_params=None):

        super().__init__(qbuildings_data, units, grids, parameters, set_indexed, cluster, method, solver, DW_params)
        self.initialize_optimization_tracking_attributes()

        # input attributes
        self.scenario = scenario.copy()
        if 'specific' not in self.scenario:
            self.scenario['specific'] = []
        if 'enforce_units' not in self.scenario:
            self.scenario['enforce_units'] = []
        if 'exclude_units' not in self.scenario:
            self.scenario['exclude_units'] = []
        if 'EMOO' not in self.scenario:
            self.scenario['EMOO'] = {}
        if 'EMOO_grid' not in self.scenario['EMOO']:
            self.scenario['EMOO']['EMOO_grid'] = 0.0
        if 'name' not in self.scenario:
            self.scenario['name'] = 'default_name'
        if 'nPareto' not in self.scenario:
            self.nPareto = 1  # no curve, single execution
            self.total_Pareto = 1
        else:
            self.nPareto = self.scenario['nPareto']  # intermediate points
            self.total_Pareto = self.nPareto * 2 + 2  # total pareto points: both objectives plus boundaries
        if self.method['building-scale'] and 'EV' not in self.infrastructure.UnitTypes:
            self.scenario['specific'] = self.scenario['specific'] + ["disallow_exchanges_1", "disallow_exchanges_2"]

        self.results = dict()

        self.solver_attributes = pd.DataFrame()
        self.epsilon_constraints = {}

    def single_optimization(self, Pareto_ID=0):
        Scn_ID = self.scenario['name']
        if self.method['district-scale'] or self.method['building-scale']:  # decomposition formulation
            ampl, exitcode = self.execute_dantzig_wolfe_decomposition(self.scenario, Scn_ID, Pareto_ID=Pareto_ID)

        else:  # compact formulation
            if self.method['use_facades'] or self.method['use_pv_orientation']:
                reho = SubProblem(self.infrastructure, self.buildings_data, self.local_data, self.parameters, self.set_indexed,
                                  self.cluster, self.scenario, self.method, self.solver, self.qbuildings_data)
            else:
                reho = SubProblem(self.infrastructure, self.buildings_data, self.local_data, self.parameters, self.set_indexed,
                                  self.cluster, self.scenario, self.method, self.solver)
            ampl = reho.build_model_without_solving()

            if self.method['fix_units']:
                for unit in self.fix_units_list:
                    for h in self.infrastructure.House:
                        if unit == 'PV':
                            ampl.getVariable('Units_Mult').get('PV_' + h).fix(self.df_fix_Units.Units_Mult.loc['PV_' + h] * 0.999)
                            ampl.getVariable('Units_Use').get('PV_' + h).fix(float(self.df_fix_Units.Units_Use.loc['PV_' + h]))
                        else:
                            ampl.getVariable('Units_Mult').get(unit + '_' + h).fix(self.df_fix_Units.Units_Mult.loc[unit + '_' + h])
                            ampl.getVariable('Units_Use').get(unit + '_' + h).fix(float(self.df_fix_Units.Units_Use.loc[unit + '_' + h]))

            ampl.solve()
            exitcode = exitcode_from_ampl(ampl)

        self.add_df_Results(ampl, Scn_ID, Pareto_ID, self.scenario)
        self.get_KPIs(Scn_ID, Pareto_ID=Pareto_ID)

        gc.collect()  # free memory
        del ampl
        if exitcode == 'infeasible':
            sys.exit(exitcode)

    def execute_dantzig_wolfe_decomposition(self, scenario, Scn_ID, Pareto_ID=0, epsilon_init=None):

        # Initiation
        self.pool = mp.Pool(mp.cpu_count())
        self.iter = 0  # new scenario has to start at iter = 0
        scenario, SP_scenario, SP_scenario_init = self.select_SP_obj_decomposition(scenario)

        self.logger.info('INITIATION, Iter:' + str(self.iter) + ' Pareto_ID: ' + str(Pareto_ID))
        self.initiate_decomposition(SP_scenario_init, Scn_ID=Scn_ID, Pareto_ID=Pareto_ID, epsilon_init=epsilon_init)
        self.logger.info('MASTER INITIATION, Iter:' + str(self.iter))
        self.MP_iteration(scenario, Scn_ID=Scn_ID, binary=False, Pareto_ID=Pareto_ID)

        # Iteration
        while self.iter < self.DW_params['max_iter'] - 1:  # last iteration is used to run the binary MP.
            self.iter += 1
            self.logger.info('SUB PROBLEM ITERATION, Iter:' + str(self.iter) + ' Pareto_ID: ' + str(Pareto_ID))
            self.SP_iteration(SP_scenario, Scn_ID=Scn_ID, Pareto_ID=Pareto_ID)
            self.logger.info('MASTER ITERATION, Iter:' + str(self.iter) + ' Pareto_ID: ' + str(Pareto_ID))
            self.MP_iteration(scenario, Scn_ID=Scn_ID, binary=False, Pareto_ID=Pareto_ID)

            if self.check_Termination_criteria(SP_scenario, Scn_ID=Scn_ID, Pareto_ID=Pareto_ID) and (self.iter > 3):
                break

        # Finalization
        self.logger.info(self.stopping_criteria)
        self.iter += 1
        self.logger.info('LAST MASTER ITERATION, Iter:' + str(self.iter) + ' Pareto_ID: ' + str(Pareto_ID))
        self.MP_iteration(scenario, Scn_ID=Scn_ID, binary=True, Pareto_ID=Pareto_ID)
        self.pool.close()

        return None, None
    
    def generate_pareto_curve(self):

        Scn_ID = self.scenario['name']

        def get_objectives_values(ampl, objectives, Pareto_ID):

            obj_values = {}
            surfaces = pd.DataFrame.from_dict({bui: self.buildings_data[bui]["ERA"] for bui in self.buildings_data}, orient="index")
            surfaces.columns = ["ERA"]

            def annualized_investment():
                if self.method['building-scale'] or self.method['district-scale']:
                    df_inv = self.results[Scn_ID][Pareto_ID]["df_Performance"]
                    district = (df_inv.Costs_inv[-1] + df_inv.Costs_rep[-1]) / self.ERA
                    buildings = df_inv.Costs_inv[:-1].div(surfaces.ERA) + df_inv.Costs_rep[:-1].div(surfaces.ERA)
                else:
                    tau = ampl.getParameter('tau').getValues().toList()  # annuality factor
                    df_h = write_results.get_ampl_data(ampl, 'Costs_House_inv', multi_index=False)
                    df1_h = write_results.get_ampl_data(ampl, 'Costs_House_rep', multi_index=False)
                    df = write_results.get_ampl_data(ampl, 'Costs_inv', multi_index=False)
                    df1 = write_results.get_ampl_data(ampl, 'Costs_rep', multi_index=False)
                    # annualized investment costs with replacements (important for BAT and NG_Cogeneration)
                    district = (df.sum()[0] + df1.sum()[0]) * tau[0] / surfaces.sum()[0]  # for compact formulation
                    buildings = (df_h.Costs_House_inv.div(surfaces.ERA) + df1_h.Costs_House_rep.div(surfaces.ERA)) * tau[0]  # for decomposition formulation
                return district, buildings

            def opex_per_house():
                if self.method['building-scale'] or self.method['district-scale']:
                    df_op = self.results[Scn_ID][Pareto_ID]["df_Performance"]
                    district = df_op.Costs_op[-1] / self.ERA
                    building = df_op.Costs_op[:-1].div(surfaces.ERA)
                else:
                    df_h = write_results.get_ampl_data(ampl, 'Costs_House_op', multi_index=False)
                    df = write_results.get_ampl_data(ampl, 'Costs_op', multi_index=False)
                    district = df.sum()[0] / surfaces.sum()[0]  # normalized OPEX CHF/m2, for compact formulation
                    building = df_h.Costs_House_op.div(surfaces.ERA)
                return district, building

            def totex_per_house():
                capex_district, capex_building = annualized_investment()
                opex_district, opex_building = opex_per_house()
                totex_district = capex_district + opex_district
                totex_house = capex_building + opex_building
                return totex_district, totex_house

            def gwp_per_house():
                df_perf = self.results[Scn_ID][Pareto_ID]["df_Performance"]
                df_GWP = (df_perf["GWP_op"] + df_perf["GWP_constr"])
                GWP_district = df_GWP["Network"] / surfaces.sum()[0]
                GWP_house = df_GWP.drop("Network").div(surfaces.ERA)
                return GWP_district, GWP_house

            for i, obj in enumerate(objectives):
                if "CAPEX" == obj:
                    obj_values["district_obj" + str(i + 1)], obj_values["building_obj" + str(i + 1)] = annualized_investment()
                elif "OPEX" == obj:
                    obj_values["district_obj" + str(i + 1)], obj_values["building_obj" + str(i + 1)] = opex_per_house()
                elif "TOTEX" == obj:
                    obj_values["district_obj" + str(i + 1)], obj_values["building_obj" + str(i + 1)] = totex_per_house()
                elif "GWP" == obj:
                    obj_values["district_obj" + str(i + 1)], obj_values["building_obj" + str(i + 1)] = gwp_per_house()
                else:
                    obj_values["district_obj" + str(i + 1)] = self.results[Scn_ID][Pareto_ID]["df_lca_Performance"][obj]["Network"] / np.sum(surfaces)[0]
                    obj_values["building_obj" + str(i + 1)] = self.results[Scn_ID][Pareto_ID]["df_lca_Performance"][obj].drop("Network").div(surfaces.ERA)
            return obj_values

        def add_constraints_from_self_scenario():
            scenario = {'EMOO': {'EMOO_grid': self.scenario['EMOO']['EMOO_grid']},
                        'specific': self.scenario['specific'],
                        'exclude_units': self.scenario['exclude_units'],
                        'enforce_units': self.scenario['enforce_units'],
                        }
            return scenario

        def find_obj1_lower_bound():
            scenario = add_constraints_from_self_scenario()
            objective1 = self.scenario["Objective"][0]
            scenario['Objective'] = objective1

            if self.method['district-scale']:
                ampl, exitcode = self.execute_dantzig_wolfe_decomposition(scenario, Scn_ID, Pareto_ID=1)
            else:
                if self.method['use_facades'] or self.method['use_pv_orientation']:
                    reho = SubProblem(self.infrastructure, self.buildings_data, self.local_data, self.parameters, self.set_indexed, self.cluster,
                                      scenario, self.method, self.solver, self.qbuildings_data)
                else:
                    reho = SubProblem(self.infrastructure, self.buildings_data, self.local_data, self.parameters, self.set_indexed, self.cluster,
                                      scenario, self.method, self.solver)
                ampl, exitcode = reho.solve_model()

            scenario = {'Objective': objective1}
            self.add_df_Results(ampl, Scn_ID, 1, scenario)
            self.get_KPIs(Scn_ID, Pareto_ID=1)

            obj_values = get_objectives_values(ampl, self.scenario["Objective"], Pareto_ID=1)

            gc.collect()  # free memory
            self.logger.info('The lower bound of the ' + str(objective1) + 'value is: ' + str(obj_values["district_obj1"]))
            return obj_values

        def find_obj2_lower_bound():
            scenario = add_constraints_from_self_scenario()
            objective2 = self.scenario["Objective"][1]
            scenario['Objective'] = objective2

            if not self.method["switch_off_second_objective"]:
                Pareto_ID = self.total_Pareto
            else:
                Pareto_ID = self.nPareto + 2

            if self.method['district-scale']:
                ampl, exitcode = self.execute_dantzig_wolfe_decomposition(scenario, Scn_ID, Pareto_ID=Pareto_ID)
            else:
                if self.method['use_facades'] or self.method['use_pv_orientation']:
                    reho = SubProblem(self.infrastructure, self.buildings_data, self.local_data, self.parameters, self.set_indexed, self.cluster,
                                      scenario, self.method, self.solver, self.qbuildings_data)
                else:
                    reho = SubProblem(self.infrastructure, self.buildings_data, self.local_data, self.parameters, self.set_indexed, self.cluster,
                                      scenario, self.method, self.solver)
                ampl, exitcode = reho.solve_model()

            scenario = {'Objective': objective2}
            self.add_df_Results(ampl, Scn_ID, Pareto_ID, scenario)
            self.get_KPIs(Scn_ID, Pareto_ID=Pareto_ID)

            obj_values = get_objectives_values(ampl, self.scenario["Objective"], Pareto_ID=Pareto_ID)

            gc.collect()  # free memory
            self.logger.info('The upper bound of the ' + str(self.scenario["Objective"][0]) + 'value is: ' + str(obj_values["district_obj1"]))
            return obj_values

        def return_epsilon_init(C_max, C_min, pareto_max, pareto, objective):
            if self.method['building-scale']:
                if objective == self.scenario["Objective"][0]:
                    epsilon_init = (C_max - C_min) / (pareto_max + 1) * (pareto - 1) + C_min
                elif objective == self.scenario["Objective"][1]:
                    epsilon_init = (C_max - C_min) / (pareto_max + 1) * (pareto + 1) + C_min
                else:
                    epsilon_init = None
            else:
                epsilon_init = None
            return epsilon_init

        def sort_pareto_points():

            df = pd.DataFrame()
            for i in self.results[Scn_ID].keys():
                if self.scenario["Objective"][0] in self.infrastructure.lca_kpis:
                    df2 = pd.DataFrame([self.results[Scn_ID][i]["df_lca_Performance"][self.scenario["Objective"][0]].xs("Network")], index=[i])
                else:
                    df2 = pd.DataFrame([self.results[Scn_ID][i]["df_Performance"]['Costs_op'].xs("Network")], index=[i])
                df = pd.concat([df, df2])
            df = df.sort_values([0], ascending=False).reset_index()

            new_order_results = {}

            rename_dict = {}
            for n, idx in enumerate(df['index'].values):
                new_order_results[n + 1] = self.results[Scn_ID][idx]
                rename_dict[idx] = n + 1

            self.results[Scn_ID] = new_order_results

            if self.method['district-scale']:
                self.sort_decomp_result(Scn_ID, df['index'].values)

        # Bounds Pareto curve
        obj1_lower_bound = find_obj1_lower_bound() 
        obj1_upper_bound = find_obj2_lower_bound()

        obj1_max = obj1_upper_bound["district_obj1"]
        obj1_min = obj1_lower_bound["district_obj1"]
        obj1_house_max = obj1_upper_bound["building_obj1"]
        obj1_house_min = obj1_lower_bound["building_obj1"]

        # Intermediate Pareto points: OPEX optimization with CAPEX constraint
        scenario = add_constraints_from_self_scenario()
        scenario['Objective'] = self.scenario["Objective"][1]
        self.epsilon_constraints['EMOO_obj1'] = np.array([])

        for nParetoIT in range(2, self.nPareto + 2):
            # Computation of the intermediate RES values
            obj1_eps_lim = (obj1_max - obj1_min) / (self.nPareto + 1) * (nParetoIT - 1) + obj1_min
            epsilon_init = return_epsilon_init(obj1_house_max, obj1_house_min, self.nPareto, nParetoIT, self.scenario["Objective"][0])

            if self.scenario["Objective"][0] in ["OPEX", "CAPEX", "TOTEX", "GWP"]:
                scenario['EMOO']['EMOO_' + self.scenario["Objective"][0]] = obj1_eps_lim
            else:
                scenario['EMOO']['EMOO_lca'] = {self.scenario["Objective"][0]: obj1_eps_lim}

            self.epsilon_constraints['EMOO_obj1'] = np.append(self.epsilon_constraints['EMOO_obj1'], obj1_eps_lim)
            self.logger.info('---------------> ' + str(self.scenario["Objective"][0]) + ' LIMIT: ' + str(obj1_eps_lim))

            # Results computation
            if self.method['district-scale']:
                ampl, exitcode = self.execute_dantzig_wolfe_decomposition(scenario, Scn_ID, Pareto_ID=nParetoIT, epsilon_init=epsilon_init)
            else:
                if self.method['use_facades'] or self.method['use_pv_orientation']:
                    reho = SubProblem(self.infrastructure, self.buildings_data, self.local_data, self.parameters, self.set_indexed,
                                      self.cluster, scenario, self.method, self.solver, self.qbuildings_data)
                else:
                    reho = SubProblem(self.infrastructure, self.buildings_data, self.local_data, self.parameters, self.set_indexed,
                                      self.cluster, scenario, self.method, self.solver)
                ampl, exitcode = reho.solve_model()

            self.add_df_Results(ampl, Scn_ID, nParetoIT, scenario)
            self.get_KPIs(Scn_ID, Pareto_ID=nParetoIT)

            del ampl
            gc.collect()  # free memory

        if not self.method['switch_off_second_objective']:

            # Intermediate Pareto points: CAPEX optimization with OPEX constraint
            scenario = add_constraints_from_self_scenario()
            scenario['Objective'] = self.scenario["Objective"][0]
            self.epsilon_constraints['EMOO_obj2'] = np.array([])

            obj2_min = obj1_upper_bound["district_obj2"]
            obj2_max = obj1_lower_bound["district_obj2"]
            obj2_house_min = obj1_upper_bound["building_obj2"]
            obj2_house_max = obj1_lower_bound["building_obj2"]

            for point, nParetoIT in enumerate(range(self.nPareto + 2, self.total_Pareto)):
                # Computation of the intermediate RES values
                obj2_eps_lim = (obj2_max - obj2_min) / (self.nPareto + 1) * (point + 1) + obj2_min
                epsilon_init = return_epsilon_init(obj2_house_max, obj2_house_min, self.nPareto, point, self.scenario["Objective"][1])

                if self.scenario["Objective"][1] in ["OPEX", "CAPEX", "TOTEX", "GWP"]:
                    scenario['EMOO']['EMOO_' + self.scenario["Objective"][1]] = obj2_eps_lim
                else:
                    scenario['EMOO']['EMOO_lca'] = {self.scenario["Objective"][1]: obj2_eps_lim}

                self.epsilon_constraints['EMOO_obj2'] = np.append(self.epsilon_constraints['EMOO_obj2'], obj2_eps_lim)
                self.logger.info('---------------> ' + str(self.scenario["Objective"][1]) + ' LIMIT: ' + str(obj2_eps_lim))
                # results computation
                if self.method['district-scale']:
                    ampl, exitcode = self.execute_dantzig_wolfe_decomposition(scenario, Scn_ID, Pareto_ID=nParetoIT, epsilon_init=epsilon_init)
                else:
                    if self.method['use_facades'] or self.method['use_pv_orientation']:
                        reho = SubProblem(self.infrastructure, self.buildings_data, self.local_data, self.parameters,
                                          self.set_indexed, self.cluster, scenario, self.method, self.solver, self.qbuildings_data)
                    else:
                        reho = SubProblem(self.infrastructure, self.buildings_data, self.local_data, self.parameters,
                                          self.set_indexed, self.cluster, scenario, self.method, self.solver)
                    ampl, exitcode = reho.solve_model()

                self.add_df_Results(ampl, Scn_ID, nParetoIT, scenario)
                self.get_KPIs(Scn_ID, Pareto_ID=nParetoIT)

                del ampl
                gc.collect()  # free memory

        sort_pareto_points()

        self.logger.info(str(obj1_min) + " " + str(obj1_max))

    def get_DHN_costs(self):

        self.pool = mp.Pool(mp.cpu_count())
        self.iter = 0  # new scenario has to start at iter = 0
        method = self.method['building-scale']
        self.method['building-scale'] = True
        scenario = self.scenario.copy()
        scenario["specific"] = scenario["specific"] + ["enforce_DHN"]
        scenario_MP, SP_scenario, SP_scenario_init = self.select_SP_obj_decomposition(scenario)

        self.initiate_decomposition(SP_scenario_init, Scn_ID=0, Pareto_ID=0)
        self.MP_iteration(scenario_MP, Scn_ID=0, binary=False, Pareto_ID=0, read_DHN=True)

        if not self.method["DHN_CO2"]:
            delta_enthalpy = np.array(self.parameters["T_DHN_supply_cst"] - self.parameters["T_DHN_return_cst"]).mean() * 4.18
        else:
            delta_enthalpy = 179.5

        f = self.feasible_solutions - 1
        heat_flow = self.results_MP[0][0][0]["df_District"]["flowrate_max"] * delta_enthalpy
        dhn_inv = self.results_MP[0][0][0]["df_District"].loc["Network", "DHN_inv"]
        tau = self.results_SP[0][0][0][f]["Building1"]["df_Performance"]["ANN_factor"][0]
        dhn_invh = dhn_inv / (tau * sum(heat_flow[0:-1]))
        for bui in self.infrastructure.houses.keys():
            self.infrastructure.Units_Parameters.loc["DHN_pipes_" + bui, ["Units_Fmax", "Cost_inv2"]] = [heat_flow[bui] * 1.001, dhn_invh]

        self.pool.close()
        self.method['building-scale'] = method
        self.initialize_optimization_tracking_attributes()

    def add_df_Results(self, ampl, Scn_ID, Pareto_ID, scenario):
        if self.method['building-scale'] or self.method['district-scale']:
            df_Results = self.get_df_Results_from_MP_and_SPs(Scn_ID, Pareto_ID)
        else:
            df_Results = write_results.get_df_Results_from_SP(ampl, scenario, self.method, self.buildings_data)
            # self.get_solver_attributes(Scn_ID, Pareto_ID, ampl)

        if Scn_ID not in self.results:
            self.results[Scn_ID] = {}
        if Pareto_ID not in self.results[Scn_ID]:
            self.results[Scn_ID][Pareto_ID] = {}

        self.results[Scn_ID][Pareto_ID] = df_Results

    def get_df_Results_from_MP_and_SPs(self, Scn_ID, Pareto_ID):

        df_Results = dict()

        # get the indexes of the SPs selected by the last MP
        last_results = self.results_MP[Scn_ID][Pareto_ID][self.iter]
        lambdas = last_results["df_DW"]['lambda']
        MP_selection = lambdas[lambdas >= 0.999].index

        # df_Time
        ids = self.number_SP_solutions.iloc[0]
        df_Time = self.results_SP[ids['Scn_ID']][ids['Pareto_ID']][ids['Iter']][ids['FeasibleSolution']][
            ids['House']]["df_Time"]

        # df_Performance
        df_Performance = self.get_final_SPs_results(MP_selection, 'df_Performance')
        df_Performance = df_Performance.groupby('Hub').sum()

        for column in ["Costs_op", "Costs_inv", "Costs_cft", "GWP_op", "GWP_constr"]:
            df_Performance.loc[:, column] = last_results["df_District"][column]
        df_Performance.loc['Network', 'ANN_factor'] = df_Performance['ANN_factor'][0]

        if self.method["actors_problem"]:
            df_actor = self.results_MP[Scn_ID][Pareto_ID][ids['Iter']]["df_District"][
                ['C_op_renters_to_utility', 'C_op_renters_to_owners', 'C_op_utility_to_owners', 'owner_inv',
                 'owner_portfolio']]
            df_Performance = pd.concat([df_Performance, df_actor], axis=1)
            df_Results["df_Actors_tariff"] = self.results_MP[Scn_ID][Pareto_ID][ids['Iter']]["df_Actors_tariff"]
            df_Results["df_Actors"] = self.results_MP[Scn_ID][Pareto_ID][ids['Iter']]["df_Actors"]

        # df_Grid_t
        df = self.get_final_SPs_results(MP_selection, 'df_Grid_t')
        df = df.droplevel(['Scn_ID', 'Pareto_ID', 'Iter', 'FeasibleSolution', 'house'])
        df = df.sort_index(level='Hub')

        h_op = df_Time.dp
        h_op.iloc[-2:] = 1
        df_network = last_results["df_District_t"].copy()
        df_network[["Network_supply", "Network_demand"]] = df_network[["Network_supply", "Network_demand"]].divide(h_op, axis=0, level='Period')

        df_network["Uncontrollable_load"] = df.groupby(["Layer", "Period", "Time"]).sum()["Uncontrollable_load"]

        df_network = pd.concat([df_network], keys=['Network'], names=['Hub']).reorder_levels(['Layer', 'Hub', 'Period', 'Time'])
        df_network = df_network.rename(columns={"Cost_demand_network": "Cost_demand",
                                                "Cost_supply_network": "Cost_supply",
                                                "Network_demand": "Grid_demand",
                                                "Network_supply": "Grid_supply"})

        columns = ["Cost_demand", "Cost_supply", "GWP_demand", "GWP_supply"]
        for h in self.buildings_data.keys():
            for column in columns:
                values_to_assign = df_network[column].values
                target_slice = df.loc[pd.IndexSlice[:, h, :, :], column]

                if len(values_to_assign) != len(target_slice):
                    raise ValueError("Mismatch between target slice length and values length")

        df_Grid_t = pd.concat([df, df_network])

        # df_Unit
        df_Unit = self.get_final_MP_results(Pareto_ID=Pareto_ID, Scn_ID=Scn_ID)
        df_Unit = df_Unit.droplevel(['FeasibleSolution', 'Hub'])
        df_Unit = df_Unit.sort_index(level='Unit')

        # df_Annuals
        df = self.get_final_SPs_results(MP_selection, 'df_Annuals')
        df = df.sort_index(level='house')
        df = df.droplevel(['Scn_ID', 'Pareto_ID', 'Iter', 'FeasibleSolution', 'house'])
        df = df.sort_index(level='Layer')
        df = df.drop('Network', level='Hub')

        df_network = pd.DataFrame(self.infrastructure.grids.keys(), columns=["Layer"])  # build a df template
        df_network["Hub"] = "Network"
        df_network = df_network.set_index(["Layer", "Hub"])
        df_network[df.columns] = float("nan")

        for key in self.infrastructure.grids.keys():
            data = df_Grid_t.xs((key, "Network"), level=("Layer", "Hub"))[["Grid_demand", "Grid_supply"]]
            data = data.mul(df_Time.dp, level='Period', axis=0)
            df_network.loc[key, ['Demand_MWh', 'Supply_MWh']] = data.sum().values / 1000

        for i, unit in enumerate(self.infrastructure.UnitsOfDistrict):
            for key in self.infrastructure.district_units[i]["UnitOfLayer"]:
                data = last_results["df_Unit_t"].xs((key, unit), level=('Layer', 'Unit')).mul(df_Time.dp, level='Period', axis=0).sum() / 1000
                df_network.loc[(key, unit), :] = float('nan')
                df_network.loc[(key, unit), ['Demand_MWh', 'Supply_MWh']] = data[['Units_demand', 'Units_supply']].values
        df_Annuals = pd.concat([df, df_network]).sort_index()

        # df_Buildings
        df_Buildings = pd.DataFrame.from_dict(self.buildings_data, orient='index')
        df_Buildings.index.names = ['Hub']
        for item in ['x', 'y', 'z', 'geometry']:
            if item in df_Buildings.columns:
                df_Buildings.drop([item], axis=1)

        if self.method['use_pv_orientation'] or self.method['use_facades']:
            # PV_Surface
            df_PV_Surface = self.get_final_SPs_results(MP_selection, 'df_PV_Surface')
            df_PV_Surface = df_PV_Surface.droplevel(['Scn_ID', 'Pareto_ID', 'Iter', 'FeasibleSolution', 'house'])
            df_PV_Surface.sort_index(level='Hub')

            # df_PV_orientation
            df_PV_orientation = self.get_final_SPs_results(MP_selection, 'df_PV_orientation')
            df_PV_orientation = df_PV_orientation.droplevel(
                ['Scn_ID', 'Pareto_ID', 'Iter', 'FeasibleSolution', 'house'])
            df_PV_orientation.sort_index(level='Hub')
            df_Results["df_PV_Surface"] = df_PV_Surface
            df_Results["df_PV_orientation"] = df_PV_orientation

        # set results
        df_Results["df_Performance"] = df_Performance
        df_Results["df_Annuals"] = df_Annuals
        df_Results["df_Unit"] = df_Unit
        df_Results["df_Grid_t"] = df_Grid_t
        df_Results["df_Time"] = df_Time

        if self.method["save_data_input"]:
            df_Results["df_Buildings"] = df_Buildings

            # df_Weather
            ids = self.number_SP_solutions.iloc[0]
            df_Weather = self.results_SP[ids["Scn_ID"]][ids["Pareto_ID"]][ids["Iter"]][ids["FeasibleSolution"]][ids["House"]]["df_Weather"]
            df_Results["df_Weather"] = df_Weather

            # df_Index
            ids = self.number_SP_solutions.iloc[0]
            df_Index = self.results_SP[ids["Scn_ID"]][ids["Pareto_ID"]][ids["Iter"]][ids["FeasibleSolution"]][ids["House"]]["df_Index"]
            df_Results["df_Index"] = df_Index

        # df_Buildings_t
        df_Buildings_t = self.get_final_SPs_results(MP_selection, 'df_Buildings_t')
        df_Buildings_t = df_Buildings_t.droplevel(['Scn_ID', 'Pareto_ID', 'Iter', 'FeasibleSolution', 'house'])
        df_Buildings_t.sort_index(level='Hub')
        df_Results["df_Buildings_t"] = df_Buildings_t

        # df_Unit_t
        df_Unit_t = self.get_final_SPs_results(MP_selection, 'df_Unit_t')
        df_Unit_t = df_Unit_t.droplevel(['Scn_ID', 'Pareto_ID', 'Iter', 'FeasibleSolution', 'house'])
        df_district_units = last_results["df_Unit_t"]
        df_Unit_t = pd.concat([df_Unit_t, df_district_units])
        df_Results["df_Unit_t"] = df_Unit_t

        if self.method["save_streams"]:
            # df_Streams_t
            df_Streams_t = self.get_final_SPs_results(MP_selection, 'df_Streams_t')
            df_Streams_t = df_Streams_t.droplevel(['Scn_ID', 'Pareto_ID', 'Iter', 'FeasibleSolution', 'house'])
            df_Results["df_Streams_t"] = df_Streams_t

        if self.method["save_lca"]:
            df_lca_Units = self.get_final_SPs_results(MP_selection, 'df_lca_Units')
            df_lca_Units = df_lca_Units.droplevel(level=["Scn_ID", "Pareto_ID", "Iter", "FeasibleSolution", "house"])
            df_Results["df_lca_Units"] = pd.concat([df_lca_Units, last_results["df_lca_Units"]]).sort_index()
            df_Results["df_lca_Performance"] = last_results["df_lca_Performance"]
            df_Results["df_lca_operation"] = last_results["df_lca_operation"]

        return df_Results

    def get_final_SPs_results(self, MP_selection, df_name):
        data = self.return_combined_SP_results(self.results_SP, df_name)
        df = pd.DataFrame()
        for idx in MP_selection.values:
            df_idx = data.xs(idx, level=('FeasibleSolution', 'house'), drop_level=False)
            df = pd.concat([df, df_idx])
        return df

    def get_KPIs(self, Scn_ID=0, Pareto_ID=0):
        df_KPI, df_Economics = calculate_KPIs(self.results[Scn_ID][Pareto_ID], self.infrastructure, self.buildings_data, self.cluster, self.local_data["df_Timestamp"], self.local_data["df_Emissions"])
        self.results[Scn_ID][Pareto_ID]["df_KPIs"] = df_KPI
        self.results[Scn_ID][Pareto_ID]["df_Economics"] = df_Economics
        if self.method['building-scale']:
            self.results[Scn_ID][Pareto_ID] = correct_network_values(self, Scn_ID, Pareto_ID)

    def save_results(self, format='pickle', filename='results', erase_file=True, filter=True):
        """
        Saves the results in the desired format: pickle file or Excel sheet.

        The results are indexed on the scenarios and pareto IDs.

        Parameters
        ----------
        format : tuple, optional
            Format(s) in which to save the results. Choose from 'pickle' and 'xlsx'.
            Default is ('pickle').
        filename : str, optional
            Base name of the file to be saved. The extension will be added based on the format.
            Default is 'results'.
        erase_file : bool, optional
            Whether to overwrite existing files with the same name.
            Default is True.
        filter : bool, optional
            Whether to filter out rows with only zeros in Excel sheets.
            Default is True.

        Returns
        -------
        None

        Notes
        -----
        If 'erase_file' is set to False, a unique counter is added to the filename to avoid overwriting existing files.

        """
        try:
            os.makedirs('results')
        except OSError:
            if not os.path.isdir('results'):
                raise

        if 'save_all' in format:
            results = self  # save the whole reho object
        else:
            results = self.results  # save only reho results

        if 'pickle' in format:
            result_file_name = str(filename) + '.pickle'
            counter = 0
            while os.path.isfile('results/' + result_file_name) and not erase_file:
                counter += 1
                result_file_name = str(filename) + '_' + str(counter) + '.pickle'

            result_file_path = 'results/' + result_file_name
            f = open(result_file_path, 'wb')
            pickle.dump(results, f)
            f.close()
            self.logger.info('Results are saved in ' + result_file_path)

        if 'xlsx' in format:

            for Scn_ID in list(results.keys()):
                for Pareto_ID in list(results[Scn_ID].keys()):

                    if Pareto_ID == 0:
                        result_file_path = 'results/' + str(filename) + '_' + str(Scn_ID) + '.xlsx'
                    else:
                        result_file_path = 'results/' + str(filename) + '_' + str(Scn_ID) + str(Pareto_ID) + '.xlsx'

                    writer = pd.ExcelWriter(result_file_path)

                    for df_name, df in results[Scn_ID][Pareto_ID].items():
                        if df is not None:
                            df = df.fillna(0)  # replace all NaN with zeros
                            if filter:
                                df = df.loc[~(df == 0).all(axis=1)]  # drop all lines with only zeros
                            df.to_excel(writer, sheet_name=df_name)

                    writer.close()
                    self.logger.info('Results are saved in ' + result_file_path)



from datetime import timedelta
from reho.model.preprocessing.sia_parser import *
from reho.model.preprocessing.QBuildings import *

__doc__ = """
Generates the buildings profiles for domestic hot water (DHW) demand, domestic electricity demand, internal heat gains, and solar gains.
"""


def reference_temperature_profile(parameters_to_ampl, cluster):
    """
    Returns a reference temperature timeseries.
    """
    # TODO: time dependent indoor temperature f.e. lower at night

    total_timesteps = cluster['Periods'] * cluster['PeriodDuration'] + 2
    T_comfort_min_0 = parameters_to_ampl['T_comfort_min_0']

    np_temperature = np.array([])
    for key in parameters_to_ampl['T_comfort_min_0']:
        np_temperature = np.append(np_temperature, np.tile(T_comfort_min_0[key], total_timesteps))

    return np_temperature


def eud_profiles(buildings_data, cluster, df_SIA_380, df_SIA_2024, df_Timestamp, include_stochasticity=False, sd_stochasticity=None, use_custom_profiles=False):
    """
    Generates building-specific profiles for internal heat gains, DHW demand, and domestic electricity demand based on
    `SIA 2024 norms <https://shop.sia.ch/collection%20des%20normes/architecte/2024_2021_f/F/Product>`_.

    The SIA profiles are daily profiles with coefficient attributed to each month.
    This function extends the profiles to the periods used, according to the building's affectation.

    Parameters
    ----------
    buildings_data : dict
        Buildings data from QBuildingsReader class.
    df_SIA_380 : pd.DataFrame
        SIA norms.
    df_SIA_2024 : pd.DataFrame
        SIA norms.
    df_Timestamp : pd.DataFrame
        Information for clustering results, used to know the periods and period duration.
    cluster : dict
        Clustering parameters.
    include_stochasticity : bool
        Includes variability in the standard values given by the SIA profiles (see :ref:`tbl-methods`).
    sd_stochasticity : list
        Parameters of the stochasticity: first value is the standard deviation on the peak demand, second value is the standard deviation on the time-shift (see :ref:`tbl-methods`).
    use_custom_profiles : dict
        Allows to give custom profiles (see :ref:`tbl-methods`).

    Returns
    -------
    np.array
        Heat gains from people
    np.array
        DHW demand
    np.array
        Electricity demand

    See also
    --------
    reho.model.preprocessing.QBuildings.QBuildingsReader

    Notes
    -----
    - One building can have several affectations. In that case, the building is divided by the share of ERA by
      affectations and the profiles are summed.
    - To use custom profiles, use csv files with 8760 rows. The name of the columns should be the same as the buildings keys in `buildings_data`.

    .. caution::

        When using custom electricity profiles, the heat gains from electricity appliances are estimated through a conversion
        factor ``conv_heat_factor`` (default value = 70%).

    Examples
    --------
    >>> my_profiles = {'electricity': 'my_folder/electricity.csv'}
    >>> file_id = 'Geneva_10_24_T_I_W'
    >>> cluster = {'Location': 'Bruxelles', 'Attributes': ['T', 'I', 'W'], 'Periods': 10, 'PeriodDuration': 24}
    >>> people_gain, eud_dhw, eud_elec = eud_profiles(buildings_data, cluster, use_custom_profiles=my_profiles)
    """
    # get cluster information
    df_Timestamp.Date = pd.to_datetime(df_Timestamp['Date'], format="%m/%d/%Y/%H")

    np_gain_all = np.array([])
    np_dhw_all = np.array([])
    np_el_all = np.array([])

    if use_custom_profiles:
        # Replace filepath in dictionary to typical profiles
        for key, val in use_custom_profiles.items():
            use_custom_profiles[key] = annual_to_typical(cluster, annual_file=val, df_Timestamp=df_Timestamp)

    for b in buildings_data:  # iterate over buildings
        # get SIA Profiles
        classes = buildings_data[b]['id_class'].split('/')
        if isinstance(buildings_data[b]['ratio'], float):
            ratios = str(buildings_data[b]['ratio'])
        else:
            ratios = buildings_data[b]['ratio'].split('/')
        status_buildings = buildings_data[b]['status'].split(',')
        np_gain_class = np.zeros(cluster['Periods'] * cluster['PeriodDuration'] + 2)
        np_dhw_class = np.zeros(cluster['Periods'] * cluster['PeriodDuration'] + 2)
        np_el_class = np.zeros(cluster['Periods'] * cluster['PeriodDuration'] + 2)
        for i, class_380 in enumerate(classes):
            # share of rooms for building type
            rooms = read_sia2024_rooms_sia380_1(class_380, df_SIA_380)
            status = ''.join(filter(str.isalnum, status_buildings[i]))
            if class_380 == 'I' or class_380 == 'II':
                area_net_floor = buildings_data[b]['ERA'] / 1.245
            else:
                area_net_floor = buildings_data[b]['ERA']
            area_net_floor = area_net_floor * float(ratios[i])
            np_gain = np.array([])
            np_dhw = np.array([])
            np_el = np.array([])

            if include_stochasticity:
                [RV_scaling, SF] = create_random_var(sd_stochasticity[0], sd_stochasticity[1])

            for p in df_Timestamp.index:  # get profiles for each typical day

                df_profiles = daily_profiles_with_monthly_deviation(status, rooms, df_Timestamp.xs(p).Date, df_SIA_2024)
                if include_stochasticity:
                    df_profiles = apply_stochasticity(df_profiles, RV_scaling, SF)

                df_custom_profiles = pd.DataFrame()
                if use_custom_profiles:
                    for key, df_key in use_custom_profiles.items():
                        df_custom_profiles[key] = df_key.xs(i + 1, level='Period').loc[:, b]
                    if include_stochasticity:
                        df_custom_profiles = apply_stochasticity(df_custom_profiles, RV_scaling, SF)

                if not df_custom_profiles.empty and 'electricity' in df_custom_profiles.columns:
                    conv_heat_factor = 0.70
                    elec_gain = conv_heat_factor * df_custom_profiles['electricity']
                    elec = df_custom_profiles['electricity']
                else:
                    elec_gain = df_profiles['elecgain_W/m2'] * area_net_floor
                    elec = df_profiles['electricity_W/m2'] * area_net_floor

                if not df_custom_profiles.empty and 'occupancy' in df_custom_profiles.columns:
                    people_gain = df_custom_profiles['occupancy']
                else:
                    people_gain = df_profiles['heatgainpeople_W/m2'] * area_net_floor

                if not df_custom_profiles.empty and 'dhw' in df_custom_profiles.columns:
                    hotwater = df_custom_profiles['dhw']
                else:
                    hotwater = df_profiles['hotwater_l/m2'] * area_net_floor

                heatgain_day = (people_gain + elec_gain) / 1000  # kW profiles for each typical day
                hot_water_day = hotwater  # L profiles for each typical day
                electric_day = elec / 1000  # kW profiles for each typical day

                if p not in df_Timestamp.index.tolist()[-2:]:
                    # sort it correctly (if first hour is not 12:00)
                    begin = df_Timestamp.xs(p).Date.hour

                    # Size = 24 hours
                    heat_day = np.concatenate((heatgain_day.iloc[begin:].values, heatgain_day.iloc[:begin].values))
                    dhw_day = np.concatenate((hot_water_day.iloc[begin:].values, hot_water_day.iloc[:begin].values))
                    el_day = np.concatenate((electric_day.iloc[begin:].values, electric_day.iloc[:begin].values))

                    # Size = PeriodDuration (= TimeEnd in ampl)
                    heat_period = np.tile(heat_day, round(cluster['PeriodDuration'] / 24))
                    dhw_period = np.tile(dhw_day, round(cluster['PeriodDuration'] / 24))
                    el_period = np.tile(el_day, round(cluster['PeriodDuration'] / 24))

                elif p == df_Timestamp.index.tolist()[-2:][0]:  # Minimum period
                    heat_period = np.array([min(heatgain_day)])
                    dhw_period = np.array([max(dhw_day)])
                    el_period = np.array([max(el_day)])
                else:  # Maximum period
                    heat_period = np.array([max(heatgain_day)])
                    dhw_period = np.array([max(dhw_day)])
                    el_period = np.array([max(el_day)])

                np_gain = np.concatenate((np_gain, heat_period))
                np_dhw = np.concatenate((np_dhw, dhw_period))
                np_el = np.concatenate((np_el, el_period))

            np_gain_class += np_gain
            np_dhw_class += np_dhw
            np_el_class += np_el

        np_gain_all = np.append(np_gain_all, np_gain_class)
        np_dhw_all = np.append(np_dhw_all, np_dhw_class)
        np_el_all = np.append(np_el_all, np_el_class)

    return np_gain_all, np_dhw_all, np_el_all


def apply_stochasticity(df_profiles, scale, SF):
    """
    Returns the daily profiles where an intensity variation (scale) and time shift factor (SF) have been applied.
    """
    # implement the intensity variation in standard profiles
    df_profiles = df_profiles * scale

    # implement the time shift in standard profiles
    shift_index = np.sign(SF) + np.int_(SF)
    shift_factor = SF - np.int_(SF)
    temp = df_profiles.to_numpy()
    temp_shifted = np.zeros((df_profiles.shape[0], df_profiles.shape[1]))
    for i in range(0, df_profiles.shape[1]):
        temp_shifted[:, i] = np.roll(temp[:, i], int(shift_index[i]), axis=0)
        if abs(shift_index[i]) > 1:
            temp[:, i] = np.roll(temp[:, i], int(shift_index[i] - np.sign(SF[i])), axis=0)
    final = temp * (1 - abs(shift_factor)) + temp_shifted * abs(shift_factor)
    df_profiles = pd.DataFrame(final, columns=df_profiles.columns, index=df_profiles.index)

    return df_profiles


def create_random_var(sd_amplitude, sd_timeshift):
    """
    Creates an array of random variables for the use of ``apply_stochasticity``.

    Notes
    -----
    The array is hard-coded to be of dimension [1,5], as it applies on the daily profiles for electricity demand, DHW demand, occupancy, electricity heat gains, and heat gains from people.

    See also
    --------
    reho.model.preprocessing.buildings_profiles.apply_stochasticity
    reho.model.preprocessing.sia_parser.daily_profiles_with_monthly_deviation
    """
    # constraints
    if sd_amplitude < 0:
        sd_amplitude = 0
    if sd_timeshift < 0:
        sd_timeshift = 0

    # create the random variable for the intensity variation
    mu = 1
    RV_scaling = np.random.normal(mu, sd_amplitude, 5)
    if RV_scaling.argmin() < 0:
        print("-------------- Negative value in the intensity variation --------------")

    # create the random variable for time-shift in standard profiles
    mu = 0
    SF = np.random.normal(mu, sd_timeshift, 5)

    return RV_scaling, SF


def annual_to_typical(cluster, annual_file, df_Timestamp, typical_file=None):
    """
    From an annual profile (8760 values), extracts the values corresponding to the typical days.
    """

    # Get which days are the typical ones
    typical_days = pd.Series([i.strftime("%m/%d/%Y") for i in df_Timestamp.Date]).values
    df_annual = file_reader(annual_file)
    t1 = pd.to_datetime('1/1/2005', dayfirst=True, infer_datetime_format=True)

    # hour 1 is between 0:00 - 1:00 and is indexed with starting hour so 0:00
    for h in df_annual.index.values:
        df_annual.loc[h, 'h'] = t1 + timedelta(hours=(int(h) - 1))

    df_annual = df_annual.set_index('h')
    df_typical = pd.DataFrame()
    for i, td in enumerate(typical_days):
        df_typical = pd.concat([df_typical, df_annual.loc[td]], sort=True)
    periods = list(range(1, len(typical_days) + 1))
    hours = list(range(1, cluster['PeriodDuration'] + 1))
    df_typical = df_typical.reset_index(drop=True)
    df_typical = df_typical.set_index(pd.MultiIndex.from_product([periods, hours], names=['Period', 'Hour']))
    if typical_file:
        df_typical.to_csv(typical_file)

    return df_typical


def solar_gains_profile(buildings_data, sia_data, local_data):
    """
    Computes the solar heat gains from the irradiance.

    It uses a typical irradiation file and uses `calc_orientation_profiles` to obtain the irradiation on the west
    facades. Additionally, the solar gains depends on the facades and on a window fraction (obtained from SIA 2024).

    Parameters
    ----------
    buildings_data : dict
        Building-specific data.
    sia_data : dict
        SIA norms.
    local_data : dict
        Location-specific data.

    Returns
    -------
    np.array
        Solar gains for each timesteps.
    """

    irr_west = local_data["df_Westfacades_irr"]

    g = np.repeat(0.5, len(irr_west))  # g-value SIA 2024
    g[irr_west > 0.2] = 0.1  # assumption that if irradiation exceeds 200 W/m2, we use sunblinds

    np_gains = np.array([])
    for b in buildings_data:
        A_facades = buildings_data[b]['area_facade_m2']
        classes = buildings_data[b]['id_class'].split('/')
        if isinstance(buildings_data[b]['ratio'], float):
            ratios = str(buildings_data[b]['ratio'])
        else:
            ratios = buildings_data[b]['ratio'].split('/')
        glass_fraction_building = 0
        for i, class_380 in enumerate(classes):
            # share of rooms for building type
            rooms = read_sia2024_rooms_sia380_1(class_380, sia_data["df_SIA_380"])
            df = sia_data["df_SIA_2024"]['data']
            glass_fraction_2024 = df['Taux de surface vitrée']
            glass_fraction_rooms = (glass_fraction_2024 * rooms).sum()
            glass_fraction_building += glass_fraction_rooms * float(ratios[i])
        gains = irr_west / 1000 * g * 0.9 * glass_fraction_building / 100 * A_facades
        # glass fraction on facades from SIA 2024, 0.9 SIA 2024: acknowledge perpendicular rays
        np_gains = np.append(np_gains, gains)

    return np_gains



# All rights reserved. ECOLE POLYTECHNIQUE FEDERALE DE LAUSANNE, Switzerland,
# IPESE Laboratory, Copyright 2021
# This work can be distributed under the Apache Software License.
# See the LICENSE file for more details.

from setuptools import setup, find_packages


def read_file(file_path):
    with open(file_path, 'r', encoding='utf-8') as file:
        return file.read()


setup(

    name='REHO',
    version='1.1.1',
    entry_points={
        'console_scripts': [
            'reho-test-import = reho.test.test_import:test_import_reho_modules',
            'reho-test-run = reho.test.test_run:test_run',
            'reho-test-plot = reho.test.test_plot:all_plots',
            'reho-download-examples = reho.test.test_examples:test_download_examples',
        ],
    },
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "amplpy",
        "ampl_module_base",
        "ampl_module_highs",
        "coloredlogs",
        "geopandas<1.0.0",
        "kaleido",
        "matplotlib",
        "numpy",
        "openpyxl",
        "pandas<2.0.0",
        "plotly",
        "psycopg2<3.0.0 ; platform_system != 'Windows'",
        "psycopg2-binary<3.0.0 ; platform_system == 'Windows'",
        "pvlib",
        "pyproj",
        "python-dotenv",
        "pytest",
        "qmcpy",
        "requests",
        "SALib",
        "scikit-learn==1.3.2",
        "scikit-learn-extra",
        "scipy",
        "setuptools",
        "shapely",
        "sqlalchemy<2.0.0",
        "urllib3",
    ],

    package_data={
        '': ['*.csv', '*.xlsx', '*.dat', '*.txt' '*.mod', '*.ini'],
    },
    author='Dorsan Lepour',
    author_email='dorsan.lepour@epfl.ch',
    maintainer="IT team of IPESE Laboratory from EPFL",
    maintainer_email="joao.ferreiradasilva@epfl.ch",
    description='Renewable Energy Hub Optimizer (REHO) - A Comprehensive Decision Support Tool for Sustainable Energy System Planning',
    long_description=read_file('README.md'),
    long_description_content_type='text/markdown',
    url='https://github.com/IPESE/REHO',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.10',
    ],
    keywords='MILP, decision support, sustainable energy systems, district optimization',
    project_urls={
        'Documentation': 'https://reho.readthedocs.io/en/main/',
        'Repository': 'https://github.com/IPESE/REHO',
        'Download': 'https://pypi.org/project/REHO',
    },
)



import pandas as pd
import numpy as np
import reho.model.preprocessing.emissions_parser as emissions
import reho.model.preprocessing.weather as weather

__doc__ = """
Calculates the KPIs resulting from the optimization.
"""


def postcompute_efficiency(df_unit, buildings_data, df_annual, df_annual_network, df_profiles, df_Weather, df_Time):
    # --------------------------------------------------------------------
    # energy
    # --------------------------------------------------------------------

    demand = df_annual['MWh_el_domestic'] + df_annual['MWh_Qsh'] + df_annual['MWh_Qdhw']
    demand_net = demand.sum()
    supply = df_annual['MWh_imp_el'] - df_annual['MWh_exp'] + df_annual['MWh_resources'] + df_annual['MWh_PV']
    supply_net = df_annual_network['MWh_el_imp'] - df_annual_network['MWh_el_exp'] + df_annual_network[
        'MWh_resources'] + df_annual['MWh_PV'].sum()

    eta_I = demand / supply
    eta_I_net = demand_net / supply_net

    # -----------------------------eta_I including PV
    # annual irradiation density
    irr_p = df_Weather['I_global'].groupby(level='Period').sum() / 1000  # kWh /m2
    irr_a = irr_p.mul(df_Time.dp, axis=0)
    irr_a = irr_a.sum() / 1000  # MWh/m2
    # reference efficiency PV panel - default 0.14 if no value as input

    PV_IRR = pd.DataFrame()
    np_m2_PV = np.array([])
    for i, h in enumerate(buildings_data):
        eff_ref = 0.14
        kWp_PV = df_unit.xs('PV_' + h)['Units_Mult']
        m2_PV = kWp_PV / eff_ref
        np_m2_PV = np.append(np_m2_PV, [m2_PV])

        MWh_IRR = pd.DataFrame([m2_PV * irr_a], columns=['MWh_IRR'], index=[h])
        PV_IRR = pd.concat([PV_IRR, MWh_IRR])
    df_PV_IRR = pd.DataFrame(index=df_annual.index)
    df_PV_IRR['MWh_IRR'] = PV_IRR['MWh_IRR'].values  # cheat to get the same index
    df_m2 = pd.DataFrame(index=df_annual.index)
    df_m2['PV'] = np_m2_PV

    supply_PV = df_annual['MWh_imp_el'] - df_annual['MWh_exp'] + df_annual['MWh_resources'] + df_PV_IRR[
        'MWh_IRR']  # bulding
    supply_PV_net = df_annual_network['MWh_el_imp'] - df_annual_network['MWh_el_exp'] + df_annual_network[
        'MWh_resources'] + df_PV_IRR['MWh_IRR'].sum()  # network
    eta_Ipv = demand / supply_PV
    eta_Ipv_net = demand_net / supply_PV_net
    # --------------------------------------------------------------------
    # exergy
    # --------------------------------------------------------------------

    LHV_ng = 50.018 / 3600  # MJ/kg -> MWh/kg [Favrat book, chapter 11, page 494 (English version)]
    ex_ng = 51.757 / 3600  # MJ/kg -> MWh/kg

    Ex_ng = (df_annual['MWh_resources'] / LHV_ng) * ex_ng

    # domestic hot water
    T_dhw = 55 + 273.15
    eta_carnot_dhw = (1 - ((df_Weather['T_ext'] + 273.15) / (T_dhw + 273.15)))
    E_dhw = df_profiles['Q_DHW'] * eta_carnot_dhw

    E_dhw_p = E_dhw.groupby(level=['Hub', 'Period']).sum()
    E_dhw_a = E_dhw_p.mul(df_Time.dp, level='Period', axis=0)
    E_dhw_a = E_dhw_a.groupby(level='Hub').sum() / 1000

    # space heating
    df_E_sh_a = pd.DataFrame()
    for house in df_profiles.index.unique(level=0):
        eta_carnot_sh = (1 - ((df_Weather['T_ext'] + 273.15) / (df_profiles['T_in'].xs(house, level=0) + 273.15)))

        E_sh = df_profiles['House_Q_heating'].xs(house, level=0) * eta_carnot_sh

        E_sh_p = E_sh.groupby(level='Period').sum()
        E_sh_a = E_sh_p.mul(df_Time.dp, axis=0)
        E_sh_a = E_sh_a.sum() / 1000
        df = pd.DataFrame(E_sh_a, index=[house], columns=['E_sh'])
        df_E_sh_a = pd.concat([df_E_sh_a, df])

    df_Exergie = pd.DataFrame(index=df_annual.index)
    df_Exergie['SH'] = df_E_sh_a.values  # cheat to get the same index
    df_Exergie['DHW'] = E_dhw_a.values
    exergie_demand = df_annual['MWh_el_domestic'] + df_Exergie['SH'] + df_Exergie['DHW']
    exergie_demand_net = exergie_demand.sum()

    exergie_supply = df_annual['MWh_imp_el'] - df_annual['MWh_exp'] + df_annual['MWh_PV'] + Ex_ng
    exergie_supply_net = df_annual_network['MWh_el_imp'] - df_annual_network['MWh_el_exp'] + df_annual[
        'MWh_PV'].sum() + Ex_ng.sum()

    eta_II = exergie_demand / exergie_supply
    eta_II_net = exergie_demand_net / exergie_supply_net

    # -----------------------------eta_II including PV

    eta_carnot_irr = (1 - ((df_Weather['T_ext'] + 273.15) / (
        6000)))  # Temperature sun 6000k - source: book: McEnvoy: Practical Handbook of Photovoltaics 2nd edition page 64
    E_irr = eta_carnot_irr * df_Weather['I_global'] / 1000  # kW/m2
    E_irr_p = E_irr.groupby(level='Period').sum()
    E_irr_a = E_irr_p.mul(df_Time.dp, axis=0)
    E_irr_a = E_irr_a.sum() / 1000  # MWh
    E_irr_a = E_irr_a * df_m2

    exergie_supply_pv = df_annual['MWh_imp_el'] - df_annual['MWh_exp'] + E_irr_a['PV'] + Ex_ng
    exergie_supply_pv_net = df_annual_network['MWh_el_imp'] - df_annual_network['MWh_el_exp'] + E_irr_a[
        'PV'].sum() + Ex_ng.sum()
    eta_IIpv = exergie_demand / exergie_supply_pv
    eta_IIpv_net = exergie_demand_net / exergie_supply_pv_net

    df_eta = pd.concat([eta_I, eta_II, eta_Ipv, eta_IIpv], axis=1)
    df_eta_net = pd.concat([eta_I_net, eta_II_net, eta_Ipv_net, eta_IIpv_net], axis=1)
    df_eta = pd.concat([df_eta, df_eta_net])
    df_eta.rename(columns={0: 'eta_I', 1: 'eta_II', 2: 'eta_Ipv', 3: 'eta_IIpv'}, inplace=True)

    return df_eta


def postcompute_security_indicators(df_annual, df_annual_network):
    df_SC = df_annual['MWh_SC'].copy()
    df_SC_net = df_annual['MWh_onsite_el'].sum() - df_annual_network['MWh_el_exp']
    df_gen = df_annual['MWh_onsite_el'].copy()

    SS = df_SC / (df_annual['MWh_imp_el'] + df_SC)
    SS_net = df_SC_net / (df_annual_network['MWh_el_imp'] + df_SC_net)

    # for no generation set SC to 1
    if (df_SC == 0).any():

        df_SC.iloc[df_SC == 0] = 1
        df_SC_net.iloc[df_SC_net == 0] = 1

        if (df_gen == 0).all():
            SC = df_SC / 1
            SC_net = df_SC_net / 1
        else:
            # one building has PV -> district has SC neq 1
            SC_net = df_SC_net / df_gen.sum()
            # make sure to set SC 1 for buildings without PV
            df_gen[df_gen == 0] = 1
            SC = df_SC / df_gen
    else:
        SC = df_SC / df_gen
        SC_net = df_SC_net / df_gen.sum()

    df = pd.concat([SC, SS], axis=1)
    net = pd.concat([SC_net, SS_net], names=[0, 1], axis=1)
    net.rename(columns={'MWh_el_exp': 0, 0: 1}, inplace=True)
    df = pd.concat([df, net])
    df.rename(columns={0: 'SC', 1: 'SS'}, inplace=True)
    return df


def postcompute_pv_penetration_curtail(df_annual, df_annual_network):
    PV_gen = df_annual['MWh_PV']
    PV_gen_net = df_annual['MWh_PV'].sum()

    E_dem = df_annual['MWh_onsite_el'] + df_annual['MWh_imp_el'] - df_annual['MWh_exp']
    E_dem_net = df_annual['MWh_onsite_el'].sum() + df_annual_network['MWh_el_imp'] - df_annual_network['MWh_el_exp']

    PVP = PV_gen / E_dem
    PVP_net = PV_gen_net / E_dem_net

    s_PVP = pd.concat([PVP, PVP_net])
    df_PVP = pd.DataFrame(s_PVP, columns=['PVP'])  # series to df with column name
    df_PVP = df_PVP

    # PV curtailmnet
    df_PVP = pd.concat([df_PVP, df_annual['PVC']], axis=1)
    df_PVP.at['Network', 'PVC'] = df_annual.PVC.sum()

    PV_gen_and_PVC = (PV_gen.to_list() + [PV_gen_net]) + df_PVP.PVC / 1000
    for i, values in enumerate(df_PVP.PVC):
        if values != 0:
            df_PVP.PVC[i] = df_PVP.PVC[i] / 1000 / PV_gen_and_PVC[i]  # MWh/MWh

    return df_PVP


def postcompute_annual_revenues(df_profiles, df_profiles_net, df_Time):
    RE_SC = df_profiles['Cost_supply'] * df_profiles['SC']
    RE_feedin = df_profiles['Cost_demand'] * df_profiles['Grid_demand']

    # # # Network we count PV consumed by neighbour as SC - assumption RE_gen - Feed-in = RE_SC !! Attention not for priceprofiles
    Cost_supply = df_profiles_net.xs('Electricity', level=0)['Cost_supply']
    Cost_demand = df_profiles_net.xs('Electricity', level=0)['Cost_demand']
    RE_gen_net = df_profiles['onsite_el'].groupby(level=['Period', 'Time']).sum()
    RE_feedin_net = df_profiles_net.xs('Electricity', level=0)['Grid_demand']

    RE_SC_net = Cost_supply * (RE_gen_net - RE_feedin_net)
    RE_feedin_net = Cost_demand * RE_feedin_net

    RE_t = RE_SC + RE_feedin
    RE_t_net = RE_SC_net + RE_feedin_net
    RE_t_net = pd.concat([RE_t_net], keys=['Network'], names=['Hub'])  # add network
    RE_t = pd.concat([RE_t, RE_t_net])

    RE_p = RE_t.groupby(level=['Hub', 'Period']).sum()
    RE_a = RE_p.mul(df_Time.dp, axis=0)
    RE_a = RE_a.groupby('Hub', level=0).sum()
    df_AR = pd.DataFrame(RE_a, columns=['AR'])

    return df_AR


def postcompute_levelized_cost_electricity(df_unit, df_annual, df_profiles, df_Time, infrastructure):
    df_LCoE = pd.DataFrame()

    C_PV_net = 0
    C_BAT_net = 0

    # LCOE1----------------------------------------------------------------------
    C_el1 = -df_profiles['Cost_demand'] * df_profiles['Grid_demand'] - df_profiles['Cost_supply'] * df_profiles['SC']
    C_el1 = C_el1.groupby(level=['Hub', 'Period']).sum()
    C_el1 = C_el1.mul(df_Time.dp, axis=0)
    C_el1 = C_el1.groupby(level=['Hub']).sum()

    # LCOE2----------------------------------------------------------------------
    C_el2 = - df_profiles['Cost_supply'] * df_profiles['Grid_supply'] + df_profiles['Cost_demand'] * df_profiles[
        'Grid_demand']

    C_el2 = C_el2.groupby(level=['Hub', 'Period']).sum()
    C_el2 = C_el2.mul(df_Time.dp, axis=0)
    C_el2 = C_el2.groupby(level=['Hub']).sum()

    # iterate for each building
    for house in infrastructure.House:
        PVPanel_name = 'default'
        Battery_name = 'default'
        for unit in infrastructure.UnitsOfHouse[house]:
            if unit in infrastructure.UnitsOfType['PV']:
                PVPanel_name = unit
            if unit in infrastructure.UnitsOfType['Battery']:
                Battery_name = unit
        C_PV = df_unit.xs(PVPanel_name)['Costs_Unit_inv']
        C_BAT = df_unit.xs(Battery_name)['Costs_Unit_inv']

        C_PV_net = C_PV_net + C_PV  # network cost
        C_BAT_net = C_BAT_net + C_BAT  # network cost battery

        kWh_PV = df_annual.loc[house]['MWh_PV'] * 1000
        kWh_house = df_annual.loc[house]['MWh_el_domestic'] * 1000

        if kWh_PV == 0:
            LCoE1 = np.nan
        else:
            LCoE1 = (C_PV + C_BAT + C_el1[house]) / kWh_PV  # CHF/kWh
        LCoE2 = (C_PV + C_BAT + C_el2[house]) / kWh_house

        df_LCoE.at[house, 'LCoE1'] = LCoE1
        df_LCoE.at[house, 'LCoE2'] = LCoE2

    # KPI on network
    kWh_house_net = df_annual.MWh_el_domestic.sum() * 1000
    kWh_PV_net = df_annual.MWh_PV.sum() * 1000

    if kWh_PV_net == 0:
        LCoE1_net = np.nan
    else:
        LCoE1_net = (C_PV_net + C_BAT_net + C_el1.sum()) / kWh_PV_net
    LCoE2_net = (C_PV_net + C_BAT_net + C_el2.sum()) / kWh_house_net

    df_LCoE.at['Network', 'LCoE1'] = LCoE1_net
    df_LCoE.at['Network', 'LCoE2'] = LCoE2_net

    return df_LCoE


def postcompute_average_emission(df_annual, df_annual_net, df_profiles, df_profiles_net, df_Time, cluster, timestamp_file, emissions_matrix):

    # Emissions
    em_supply_dy = df_profiles_net.GWP_supply.xs('Electricity')
    em_demand_dy = df_profiles_net.GWP_demand.xs('Electricity')

    # Buildings
    em_el_av_bui = em_supply_dy.mean() * df_profiles.Grid_supply - em_demand_dy.mean() * df_profiles.Grid_demand
    em_el_av_bui = em_el_av_bui.mul(df_Time.dp, level='Period', axis=0).groupby('Hub').sum()

    em_el_dy_bui = pd.Series(dtype='float')
    for h in df_annual.index.get_level_values(level='Hub'):
        em_el_dy_bui_h = em_supply_dy * df_profiles.Grid_supply.xs(h,
                                                                   level='Hub') - em_demand_dy * df_profiles.Grid_demand.xs(
            h, level='Hub')
        em_el_dy_bui_h = em_el_dy_bui_h.mul(df_Time.dp, level='Period', axis=0).sum()
        em_el_dy_bui_h = pd.DataFrame([em_el_dy_bui_h], index=[h])
        em_el_dy_bui = pd.concat([em_el_dy_bui, em_el_dy_bui_h])

    # Network
    em_el_av_net = em_supply_dy.mean() * df_profiles_net.Grid_supply.xs(
        'Electricity') - em_demand_dy.mean() * df_profiles_net.Grid_demand.xs('Electricity')
    em_el_av_net = em_el_av_net.mul(df_Time.dp, level='Period', axis=0).sum()
    em_el_av_net = pd.DataFrame([em_el_av_net], index=['Network'])

    em_el_dy_net = em_supply_dy * df_profiles_net.Grid_supply.xs(
        'Electricity') - em_demand_dy * df_profiles_net.Grid_demand.xs('Electricity')
    em_el_dy_net = em_el_dy_net.mul(df_Time.dp, level='Period', axis=0).sum()
    em_el_dy_net = pd.DataFrame([em_el_dy_net], index=['Network'])

    em_el_dy = pd.concat([em_el_dy_bui, em_el_dy_net])
    em_el_av = pd.concat([em_el_av_bui, em_el_av_net])

    # --------------------------------------------------------------------
    # Renewable energy share
    # --------------------------------------------------------------------
    df_el_net = df_profiles_net.xs('Electricity', level=0)

    File_ID = weather.get_cluster_file_ID(cluster)
    res_profile = emissions.return_typical_emission_profiles(df_Time, File_ID, 'method 1', timestamp_file, emissions_matrix)
    res_av = emissions.find_average_value('CH', 'method 1', emissions_matrix)
    s_RES_dy = pd.Series(dtype='float')
    s_RES_av = pd.Series(dtype='float')

    for h in df_annual.index.get_level_values(level='Hub'):
        res_e = res_profile['GWP_supply'].values * df_profiles.Grid_supply.xs(h, level='Hub', drop_level=False)
        res_e = res_e.groupby(level=['Hub', 'Period']).sum()
        res_e = res_e.mul(df_Time.dp, axis=0).groupby(
            level='Hub').sum() / 1000  # annual emissions from elec with dy profiles ton/year
        res_av.index = res_e.index

        df_h = df_annual.loc[h]
        RES_dy = (df_h['MWh_SC'] + res_e) / (df_h['MWh_SC'] + df_h['MWh_resources'] + df_h['MWh_imp_el'])
        RES_av = (df_h['MWh_SC'] + res_av * df_h['MWh_imp_el']) / (
                    df_h['MWh_SC'] + df_h['MWh_resources'] + df_h['MWh_imp_el'])

        RES_dy = RES_dy
        RES_av = RES_av
        s_RES_dy = pd.concat([s_RES_dy, RES_dy])
        s_RES_av = pd.concat([s_RES_av, RES_av])

    # Network
    res_e = res_profile['GWP_supply'].values * df_el_net.Grid_supply
    res_e = res_e.groupby(level=['Period']).sum()
    res_e = res_e.mul(df_Time.dp, axis=0).sum() / 1000
    RES_dy = (df_annual['MWh_SC'].sum() + res_e) / (
                df_annual['MWh_SC'].sum() + df_annual_net['MWh_resources'] + df_annual_net['MWh_el_imp'])
    RES_av = (df_annual['MWh_SC'].sum() + res_av.values[0] * df_annual_net['MWh_el_imp']) / (
                df_annual['MWh_SC'].sum() + df_annual_net['MWh_resources'] + df_annual_net['MWh_el_imp'])

    RES_dy = RES_dy
    RES_av = pd.DataFrame(RES_av)
    s_RES_dy = pd.concat([s_RES_dy, RES_dy])
    s_RES_av = pd.concat([s_RES_av, RES_av])

    df = pd.concat([em_el_av, em_el_dy, s_RES_dy, s_RES_av], axis=1)
    df.columns = ['gwp_elec_av', 'gwp_elec_dy', 'RES_dy', 'RES_av']
    return df


def postcompute_Grid_param(df_Grid):
    df = df_Grid.xs('Electricity', level=0)[['Grid_demand', 'Grid_supply']]
    df_max = df.groupby(level=['Hub', 'Period']).max()
    df_mean = df.groupby(level=['Hub', 'Period']).mean().replace(0,
                                                                 1)  # replace 0 with 1 to avoid div by 0,  profiles >0, in case av = 0- whole profile is 0
    GM = df_max.div(df_mean, level='Hub').groupby('Hub').max()

    GM = GM.rename(columns={'Grid_supply': 'GMs', 'Grid_demand': 'GMd'})
    GM['GUs'] = GM.GMs
    GM['GUd'] = GM.GMd
    uncontrollable_load = df_Grid.xs('Electricity', level='Layer')['Uncontrollable_load'].drop('Network', level='Hub')
    uncontrollable_load = uncontrollable_load.groupby(level=['Period', 'Time']).sum().max()
    for h in GM.index:
        GUs = df_Grid.xs(('Electricity', h), level=('Layer', 'Hub'))['Grid_supply'][:-2].max() / uncontrollable_load
        GUd = df_Grid.xs(('Electricity', h), level=('Layer', 'Hub'))['Grid_demand'][:-2].max() / uncontrollable_load
        GM.at[h, 'GUs'] = GUs
        GM.at[h, 'GUd'] = GUd

    return GM.round(2)


def postcompute_annual_COP(df_annuals, infrastructure):
    df = pd.DataFrame(index=['Network'])
    total_heat_network = 0
    total_HP_el = 0
    for HP in infrastructure.UnitsOfType['HeatPump']:
        # get values for each AW in the district
        df_annuals_HP = df_annuals.xs(HP, level='Hub')

        df_el_all = df_annuals_HP.xs('Electricity')
        df_heat_all = df_annuals_HP.drop('Electricity').sum()
        if df_el_all.Demand_MWh == 0:
            COP = np.nan
        else:
            COP = df_heat_all['Supply_MWh'] / df_el_all.Demand_MWh

        # assign to house and df
        for key, units in infrastructure.UnitsOfHouse.items():
            if HP in units and key not in df.index:
                df.at[key, 'COP'] = COP

        # sum for network average
        total_heat_network = total_heat_network + df_heat_all['Supply_MWh']
        total_HP_el = total_HP_el + df_el_all.Demand_MWh
    if total_HP_el == 0:
        df.at['Network', 'COP'] = np.nan
    else:
        df.at['Network', 'COP'] = total_heat_network / total_HP_el

    return df


def build_df_profiles_house(df_Results, infrastructure):
    """
    Builds hourly profiles for demand and consumption of units and buildings.
    """

    df_PV = units_power_profiles_per_building(df_Results, infrastructure, 'PV')
    df_BAT = units_power_profiles_per_building(df_Results, infrastructure, 'Battery')

    df_grid_profile = df_Results["df_Grid_t"].xs('Electricity', level='Layer')

    df_profiles_house = df_grid_profile.drop('Network', level='Hub')
    df_profiles_house['PV'] = df_PV['Units_supply']
    df_profiles_house['PVC'] = df_PV['Units_curtailment']
    df_profiles_house['BA_in'] = df_BAT['Units_demand']
    df_profiles_house['BA_out'] = df_BAT['Units_supply']
    df_profiles_house['House_Q_heating'] = df_Results["df_Buildings_t"]['House_Q_heating']
    df_profiles_house['T_in'] = df_Results["df_Buildings_t"]['T_in']
    df_profiles_house['Q_DHW'] = df_Results["df_Buildings_t"]['House_Q_DHW']

    if 'NG_Cogeneration' in infrastructure.UnitTypes and not any(infrastructure.UnitsOfType["NG_Cogeneration"] == 'NG_Cogeneration_district'):
        df_NG_Cogeneration = units_power_profiles_per_building(df_Results, infrastructure, 'NG_Cogeneration')
        df_profiles_house['NG_Cogeneration'] = df_NG_Cogeneration['Units_supply']
        df_profiles_house['onsite_el'] = df_profiles_house['PV'] + df_profiles_house['NG_Cogeneration']
        PV_sell = df_profiles_house['Grid_demand'] - df_profiles_house['NG_Cogeneration']
        PV_sell[PV_sell < 0] = 0
        df_profiles_house['PV_SC'] = df_profiles_house['PV'] - PV_sell
    else:
        df_profiles_house['onsite_el'] = df_profiles_house['PV']
        df_profiles_house['PV_SC'] = df_profiles_house['onsite_el'] - df_profiles_house['Grid_demand']
    df_profiles_house['SC'] = df_profiles_house['onsite_el'] - df_profiles_house['Grid_demand']

    return df_profiles_house


def build_df_annual(df_Results, df_profiles_house, infrastructure, df_Time):
    """
    Transforms profiles to annual values, convert to MWh and insert additional values (costs, net resource exchanges).

    Parameters:
        df_Results (df) : results of a scenario
        df_profiles_house (df)
        infrastructure (df)
        df_Time (df)

    Returns:
        Annual parameters for each building and for the network.
    """

    df_period = df_profiles_house.groupby(level=['Hub', 'Period']).sum()  # 'daily' sum
    df_period = df_period.mul(df_Time.dp, level='Period', axis=0)  # multiply by frequency

    df_annual = df_period.groupby(level=['Hub']).sum()  # annuals
    df_annual['Cost_supply'] = df_annual['Cost_supply'] / 8760  # average price
    df_annual['Cost_demand'] = df_annual['Cost_demand'] / 8760

    # pass from kWh to MWh
    keys = ['Grid_supply', 'Grid_demand', 'PV', 'Q_DHW', 'House_Q_heating', 'BA_out', 'BA_in', 'onsite_el', 'SC',
            'PV_SC']
    for key in keys:
        df_annual[key] = df_annual[key].div(1000)

    df_annual = df_annual.drop(columns=['T_in', 'GWP_demand', 'GWP_supply'])

    df_annual.rename(columns={'BA_out': 'MWh_BAT_out', 'BA_in': 'MWh_BAT_in',
                              'onsite_el': 'MWh_onsite_el', 'PV': 'MWh_PV', 'Grid_demand': 'MWh_exp',
                              'Grid_supply': 'MWh_imp_el',
                              'House_Q_heating': 'MWh_Qsh', 'Q_DHW': 'MWh_Qdhw', 'SC': 'MWh_SC', 'PV_SC': 'MWh_PV_SC'},
                     inplace=True)

    # get Annual values from ampl results
    df_power = df_Results["df_Annuals"].xs('Electricity', level=0)
    df_Building_el = df_power.loc[df_annual.index]  # index: all houses
    df_annual['MWh_el_domestic'] = df_Building_el['Demand_MWh']

    df_annual_network = pd.DataFrame(index=['Network'])
    for resource in infrastructure.grids.keys():
        if resource == 'Electricity':
            df_annual_network['MWh_el_exp'] = df_power.xs('Network')['Demand_MWh']
            df_annual_network['MWh_el_imp'] = df_power.xs('Network')['Supply_MWh']
        else:
            df_resource = df_Results["df_Annuals"].xs(resource, level=0)
            df_annual_network['MWh_' + resource] = df_resource.xs('Network')['Supply_MWh']
            df_resource_houses = split_units_to_buildings(infrastructure, df_resource, 'Demand_MWh')
            df_annual['MWh_' + resource] = df_resource_houses

    resource_list = list(
        x for x in infrastructure.grids if x not in ['Electricity'])  # all resources except electricity
    df_annual['MWh_resources'] = sum([df_annual['MWh_' + x] for x in resource_list])
    df_annual_network['MWh_resources'] = sum([df_annual_network['MWh_' + x] for x in resource_list])

    df_Building_performance = df_Results["df_Performance"].loc[df_annual.index]
    df_annual['Costs_House_op'] = df_Building_performance['Costs_op']
    df_annual['Costs_House_inv'] = df_Building_performance['Costs_inv']
    df_annual['Costs_House_rep'] = df_Building_performance['Costs_rep']

    for key in ['Costs_op', 'Costs_inv', 'Costs_rep']:
        df_annual_network[key] = df_Results["df_Performance"].iloc[0][key]

    return df_annual, df_annual_network


def calculate_KPIs(df_Results, infrastructure, buildings_data, cluster, timestamp_file, emissions_matrix):
    df_profiles = build_df_profiles_house(df_Results, infrastructure)
    df_profiles_network = df_Results["df_Grid_t"].xs('Network', level='Hub').copy()

    df_Time = df_Results["df_Time"]
    df_Time.dp.iloc[-1] = 0  # exclude extreme periods
    df_Time.dp.iloc[-2] = 0

    # ------------------------------------------------------------------------------------------------------
    # Construct Economics dataframe
    # ------------------------------------------------------------------------------------------------------
    df_Economics = build_df_Economics(df_Results, df_profiles)
    df_annual, df_annual_network = build_df_annual(df_Results, df_profiles, infrastructure, df_Time)

    # ------------------------------------------------------------------------------------------------------
    # Construct KPI-Dataframe
    # ------------------------------------------------------------------------------------------------------
    df_KPI = pd.DataFrame()
    # get heated surface for normalization
    df_hsA = pd.DataFrame()
    for h in buildings_data:
        df = pd.DataFrame([buildings_data[h]['ERA']], columns=['ERA'], index=[h])
        df_hsA = pd.concat([df_hsA, df])
    network = pd.DataFrame(df_hsA.sum().values, index=['Network'], columns=['ERA'])
    df_hsA = pd.concat([df_hsA, network])

    # ------------------------------------------------------------------------------------------------------
    # Costs
    # ------------------------------------------------------------------------------------------------------
    df_cost = df_Results["df_Performance"][['Costs_op', 'Costs_inv', 'Costs_rep', 'Costs_ft']]
    df_cost = df_cost.rename(
        columns={'Costs_op': 'opex_m2', 'Costs_inv': 'capex_m2', 'Costs_rep': 'cost_rep_m2', 'Costs_ft': 'cost_ft_m2'})
    df_KPI = pd.concat([df_KPI, df_cost.div(df_hsA.ERA, axis=0)], axis=1)  # [CHF/m2/yr]

    df_AR = postcompute_annual_revenues(df_profiles, df_profiles_network, df_Time)
    df_KPI = pd.concat([df_KPI, df_AR.div(df_hsA.ERA, axis=0)], axis=1)

    df_LCoE = postcompute_levelized_cost_electricity(df_Results["df_Unit"], df_annual, df_profiles, df_Time,
                                                     infrastructure)
    df_KPI['LCoE1'] = df_LCoE.LCoE1
    df_KPI['LCoE2'] = df_LCoE.LCoE2

    # ------------------------------------------------------------------------------------------------------
    # PV related KPIs
    # ------------------------------------------------------------------------------------------------------
    df_SC = postcompute_security_indicators(df_annual, df_annual_network)
    df_KPI = pd.concat([df_KPI, df_SC], axis=1)  # SC  SS

    df_PVP = postcompute_pv_penetration_curtail(df_annual, df_annual_network)
    df_KPI = pd.concat([df_KPI, df_PVP], axis=1)  # PVP [-] PVC [MWh]

    df_GM = postcompute_Grid_param(df_Results["df_Grid_t"])
    df_KPI = pd.concat([df_KPI, df_GM], axis=1)  # GMd GMs GUd GUs

    # ------------------------------------------------------------------------------------------------------
    # GWP
    # ------------------------------------------------------------------------------------------------------
    df_KPI['gwp_op_m2'] = df_Results["df_Performance"]['GWP_op'].div(df_hsA.ERA)
    df_KPI['gwp_constr_m2'] = df_Results["df_Performance"]['GWP_constr'].div(df_hsA.ERA)
    df_KPI['gwp_tot_m2'] = df_KPI['gwp_op_m2'] + df_KPI['gwp_constr_m2']  # [kgCO2-eq/m2/yr]

    df_G_RES = postcompute_average_emission(df_annual, df_annual_network, df_profiles, df_profiles_network, df_Time, cluster, timestamp_file, emissions_matrix)
    df_KPI = pd.concat([df_KPI, df_G_RES[['gwp_elec_av', 'gwp_elec_dy']].div(df_hsA.ERA, axis=0)], axis=1)
    df_KPI = df_KPI.rename(
        columns={'gwp_elec_av': 'gwp_elec_av_m2', 'gwp_elec_dy': 'gwp_elec_dy_m2'})  # gwp_elec_av_m2    gwp_elec_dy_m2
    df_KPI = pd.concat([df_KPI, df_G_RES[['RES_dy', 'RES_av']]], axis=1)  # RES_dy    RES_av

    # ------------------------------------------------------------------------------------------------------
    # Technical KPIs
    # ------------------------------------------------------------------------------------------------------
    #df_eta = postcompute_efficiency(df_Results["df_Unit"], buildings_data, df_annual, df_annual_network, df_profiles,
    #                                df_Results["df_Weather"], df_Time)
    #df_KPI = pd.concat([df_KPI, df_eta], axis=1)  # eta_I    eta_II   eta_Ipv  eta_IIpv

    if 'HeatPump' in infrastructure.UnitsOfType:  # Check if HP DHN is used
        df_COP = postcompute_annual_COP(df_Results["df_Annuals"], infrastructure)
        df_KPI = pd.concat([df_KPI, df_COP], axis=1)

    df_KPI.index.names = ['Hub']

    return df_KPI, df_Economics


def split_units_to_buildings(infrastructure, df, aim):
    df_h = pd.DataFrame(index=infrastructure.House, columns=[aim])
    for h in infrastructure.House:
        value = np.array([])
        for idx in df.index:
            unit = idx.split('_')
            if h in unit:
                value = np.append(value, df.loc[idx, aim])
        df_h.loc[h, aim] = value.sum()
    return df_h


def units_power_profiles_per_building(df_Results, infrastructure, unittype):
    df = pd.DataFrame()
    for house in infrastructure.House:
        for unit in infrastructure.UnitsOfType[unittype]:
            if unit in infrastructure.UnitsOfHouse[house]:
                df_profile = df_Results["df_Unit_t"].xs(('Electricity', unit), level=('Layer', 'Unit'))
                df_profile = pd.concat([df_profile], keys=[(house, unit)], names=['Hub', 'Unit'])
                df = pd.concat([df, df_profile])
    df = df.groupby(level=['Hub', 'Period', 'Time']).sum()
    return df


def remove_building_from_index(df):
    def filter_building_str(str):
        if 'Building' in str:
            str_split = str.split("_")
            if len(str_split) > 2:
                new_idx = "_".join(str.split("_", 2)[:2])

            else:
                new_idx = str_split[0]
            hub = str_split[-1]
        else:
            new_idx = str
            hub = 'Network'
        return new_idx, hub

    new_index = []
    index_frame = df.index.to_frame()
    if 'Unit' in index_frame.columns:
        for idx in index_frame['Unit']:
            new_index.append(filter_building_str(idx))
        new_index = pd.DataFrame.from_records(new_index, columns=['Unit', 'Hub'])
        index_frame[['Unit', 'Hub']] = new_index.values
    elif 'Hub' in index_frame.columns:
        for idx in index_frame['Hub']:
            new_index.append(filter_building_str(idx)[0])
        index_frame['Hub'] = new_index

    index_modified = pd.MultiIndex.from_frame(index_frame)

    return df.set_index(index_modified)


def build_df_Economics(df_Results, df_profiles):
    period_duration = df_Results["df_Time"].dp
    df_unit_t = remove_building_from_index(df_Results["df_Unit_t"])
    df_unit = remove_building_from_index(df_Results["df_Unit"])
    df_grid_t = df_Results["df_Grid_t"]
    hubs = df_grid_t.index.get_level_values('Hub').unique().tolist()
    idx = pd.Index(hubs, name='Hub')

    col_to_calc = ['PV', 'PVC', 'BA_out', 'onsite_el', 'PV_SC', 'SC']
    # Avoided from profiles
    df_cost = pd.DataFrame()
    df_impact = pd.DataFrame()
    for col in col_to_calc:
        df_cost['avoided_' + col] = df_profiles.loc[:, col] * df_profiles.loc[:, 'Cost_demand']
        df_impact['avoided_' + col] = df_profiles.loc[:, col] * df_profiles.loc[:, 'GWP_demand']

    df_cost = df_cost.mul(period_duration, level='Period', axis=0).groupby(level='Hub').sum()
    df_impact = df_impact.mul(period_duration, level='Period', axis=0).groupby(level='Hub').sum()
    df_cost.loc['Network', :] = df_cost.sum()
    df_impact.loc['Network', :] = df_impact.sum()

    # Grid
    df_grid_cost = pd.DataFrame()
    df_grid_impact = pd.DataFrame()
    df_grid_cost['price_demand'] = df_grid_t.Cost_demand * df_grid_t.Grid_demand
    df_grid_cost['price_supply'] = df_grid_t.Cost_supply * df_grid_t.Grid_supply
    df_grid_impact['impact_demand'] = df_grid_t.GWP_demand * df_grid_t.Grid_demand
    df_grid_impact['impact_supply'] = df_grid_t.GWP_supply * df_grid_t.Grid_supply

    curtailment = df_unit_t['Units_curtailment'].groupby(['Layer', 'Period', 'Time', 'Hub']).sum()
    df_grid_cost['price_curtailment'] = curtailment * df_grid_t.Cost_demand
    df_grid_impact['impact_curtailment'] = curtailment * df_grid_t.GWP_demand

    df_grid_cost = df_grid_cost.mul(period_duration, level='Period', axis=0).groupby(level=['Hub', 'Layer']).sum()
    df_grid_impact = df_grid_impact.mul(period_duration, level='Period', axis=0).groupby(level=['Hub', 'Layer']).sum()

    # Unit
    df_unit = df_unit.groupby(level=['Unit', 'Hub']).sum()
    df_unit_cost = df_unit.reset_index().pivot(index='Hub', values='Costs_Unit_inv', columns='Unit')
    df_unit_impact = df_unit.reset_index().pivot(index='Hub', values='GWP_Unit_constr', columns='Unit')
    df_unit_cost.loc['Network', :] = df_unit_cost.sum()
    df_unit_impact.loc['Network', :] = df_unit_impact.sum()

    dict_to_keep = {'demand': 'revenues_', 'supply': 'costs_', 'curtailment': 'curtailment_'}
    pivot_df_grid_cost = pd.DataFrame(index=idx)
    pivot_df_grid_impact = pd.DataFrame(index=idx)
    for col, val in dict_to_keep.items():
        cost = df_grid_cost.reset_index().pivot(index='Hub', columns='Layer', values='price_' + col)
        impact = df_grid_impact.reset_index().pivot(index='Hub', columns='Layer', values='impact_' + col)
        rename_col = {key: val + key for key in cost.columns}
        cost = cost.rename(columns=rename_col)
        impact = impact.rename(columns=rename_col)
        pivot_df_grid_cost = pivot_df_grid_cost.merge(cost, on='Hub')
        pivot_df_grid_impact = pivot_df_grid_impact.merge(impact, on='Hub')

    # Final dfs
    df_op = pd.concat([df_cost.merge(pivot_df_grid_cost, on='Hub')], keys=['costs'], names=['Perf_type'])
    df_op_impact = pd.concat([df_impact.merge(pivot_df_grid_impact, on='Hub')], keys=['impact'], names=['Perf_type'])
    df_op = pd.concat([df_op, df_op_impact])
    df_inv = pd.concat([df_unit_cost, df_unit_impact], keys=['costs', 'impact'], names=['Perf_type'])

    df_Economics = pd.concat([df_op, df_inv], keys=['operation', 'investment'], names=['Category'], axis=1)

    return df_Economics


def temperature_profile(df_Results, daily_averaging=False):
    """
    Returns a pd.Series of the indoor temperature profile, one column per building.

    Parameters:
        df_Results (df) : pd.DataFrame of a scenario
        daily_averaging (bool) : to average over days

    Returns:
        df_Tin
    """

    # TODO: check if multi - buildings works fine

    # Extract the data
    buildings_list = list(df_Results['df_Buildings'].index)
    df_buildings_t = df_Results['df_Buildings_t']

    # Prepare the df to store the Tin
    df_Tin = pd.DataFrame()

    # list of the two last periods used for design purpose to drop later
    period_to_drop = list(df_buildings_t.xs('Building1').index.get_level_values('Period').unique()[-2:])

    # For each building: calculation of the Tin series
    for b in buildings_list:
        # Take Tin for the building in df_buildings_t
        df_Tin_building = pd.DataFrame()
        df_Tin_building['T_in'] = df_buildings_t.xs(b)['T_in']

        # drop the last two periods for design
        for i in period_to_drop:
            df_Tin_building = df_Tin_building.drop(i, level='Period')

        # array to work for a building
        Tin_building = np.array([])

        for j in range(1, 366):
            id = df_Results['df_Index'].PeriodOfYear[j * 24]
            data_id = df_Tin_building.xs(id)
            Tin_building = np.concatenate((Tin_building, data_id['T_in']))

        if daily_averaging:
            items_average = 24
            Tin_building = np.mean(Tin_building.reshape(-1, items_average), axis=1)

        # Store the data in df_Tin (one col per building)
        df_Tin['T_in_' + str(b)] = Tin_building

    # create idx
    idx = list(df_Tin.index + 1)

    return df_Tin.to_numpy(), idx


import pytest
from reho.model.preprocessing.clustering import Clustering
from reho.model.preprocessing.weather import get_weather_data


@pytest.fixture
def qbuildings_data():
    return {'buildings_data': {'Building1': {'x': 2496193, 'y': 1114279, 'z': 402}}}


@pytest.fixture
def weather_data(qbuildings_data):
    return get_weather_data(qbuildings_data).reset_index(drop=True)


@pytest.fixture
def cluster():
    return {'Location': 'Geneva', 'Attributes': ['T', 'I', 'W'], 'Periods': 10, 'PeriodDuration': 24}


def test_clustering(qbuildings_data, weather_data, cluster):
    cl = Clustering(data=weather_data, nb_clusters=[cluster['Periods']], option={"year-to-day": True, "extreme": []}, pd=cluster['PeriodDuration'])
    cl.run_clustering()
